/*
	proute.c is based on
	xroute.c, Copyright (c) 1993, 1994, by Kartik Subbarao

	Just a dirty hack by Dobrica Pavlinusic, 1996.
	1997-10-23 DbP	added a bind to IP adress (-b)
	1997-10-24 DbP	added repeat options (-n and -f)

 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <string.h>
#include <sys/wait.h>
#include <values.h>
#ifdef sgi
#define _BSD_SIGNALS
#endif
#include <signal.h>
#include <errno.h>

extern char *sys_errlist[]; extern int errno; 

#define XPORT 1024	/* lowest port to listen to */

usage()
{
	fprintf(stderr, "usage: proute [opts ...] host[:port]\n");
	fprintf(stderr, "opts:	-l port	port to listen for incomming connections\n");
	fprintf(stderr, "	-b IP	bind only to one IP adress\n");
	fprintf(stderr, "	-n num	allow only num connections (default: 1)\n");
	fprintf(stderr, "	-f	connect forever (overrides -n)\n");
	exit(1);
}

#if defined(hpux) || defined(linux) || defined(__osf__)
waiter() {int st; while (wait3(&st,WNOHANG,NULL) > 0); signal(SIGCHLD,waiter);}
#else
waiter() { union wait st; while (wait3(&st, WNOHANG, NULL) > 0); }
#endif

/* administrative limits */
int stdisp = 0, endisp = MAXINT-XPORT, maxcli = MAXINT; 

unsigned int ipaddr = 0; /* IP adress to bind to, 0=all on host */

main(argc, argv)
int argc; char *argv[];
{
	int o; extern char *optarg; extern int optind;
	char *xhost, *p;
	int xdisp, gdisp = -1; /* Display numbers */
	int xs, gs, os; /* X socket, gateway socket, other socket */
	int numcli = 1; /* Number of clients */
	int dec = 1; /* decrement value for numcli */

opts:
	optind = 1;
	while ((o = getopt(argc, argv, "l:b:n:f")) != -1)
		switch (o) {
		case 'l':
			gdisp = atoi(optarg); 
			break;
		case 'b':
			ipaddr = inet_addr(optarg);
			break;
		case 'n':
			numcli = atoi(optarg); 
			break;
		case 'f':
			dec = 0; 
			break;
		default: usage();
		}

	if (!(xhost = argv[optind])) usage();

	xdisp = (p = strchr(xhost, ':')) && p[1] ? atoi(p+1) : 23;	/* 23 = telnet */
	if (p) *p = '\0';

	gs = xsocket(gdisp);
	signal(SIGCHLD, waiter);
	signal(SIGHUP, SIG_IGN);

	if (fork()) exit(0); /* background ourself */

	while (numcli) {
		struct sockaddr_in rad; int radlen = sizeof rad;

		while ((os = accept(gs, &rad, &radlen)) == -1 && errno == EINTR);
		if (os == -1) perror("accept"), exit(1);

		switch (fork()) {
		case -1: exit(1);
		case 0: 
			close(gs); 
			xs = tcpcsocket(xhost, xdisp); route(xs, os); 
			break;
		default: close(os);
		}
		numcli = numcli - dec;
	}
	exit(0);
}

route(xs, os)
int xs, os;
{
	int maxfds, xn = 0, on = 0, tmp; 
	fd_set rdset, wrset, saveset;
	char xobuf[4096], *xop, oxbuf[4096], *oxp;
	
	maxfds = (xs > os ? xs : os) + 1;
	FD_ZERO(&saveset); FD_SET(xs, &saveset); FD_SET(os, &saveset);
	xop = xobuf; oxp = oxbuf;

	for (;;) {
		if (xn == 0 && on == 0) {
			rdset = saveset;
			if (select(maxfds, &rdset, NULL, NULL, NULL) == -1) exit(1);
			if (FD_ISSET(xs, &rdset))
				if ((xn = read(xs, xobuf, sizeof xobuf)) <= 0) exit(xn);
			if (FD_ISSET(os, &rdset))
				if ((on = read(os, oxbuf, sizeof oxbuf)) <= 0) exit(on);
		}
		else {
			if (xn) FD_SET(xs, &rdset);
			if (on) FD_SET(os, &rdset);
		}
	
		wrset = saveset; 
		if (select(maxfds, NULL, &wrset, NULL, NULL) == -1) exit(1);

		if (FD_ISSET(xs, &rdset) && FD_ISSET(os, &wrset)) {
			if ((tmp = write(os, xop, xn)) <= 0) exit(tmp);
			if (xn -= tmp) xop += tmp; 
			else { FD_CLR(xs, &rdset); xop = xobuf; xn = 0; }
		}

		if (FD_ISSET(os, &rdset) && FD_ISSET(xs, &wrset)) {
			if ((tmp = write(xs, oxp, on)) <= 0) exit(tmp);
			if (on -= tmp) oxp += tmp;
			else { FD_CLR(os, &rdset); oxp = oxbuf; on = 0; }
		}
	}
}

int xsocket(gdisp)
int gdisp;
{
	int i, s; struct sockaddr_in sa;
   	 
    if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		perror("socket"), exit(1);

   	sa.sin_family = AF_INET; 
   	
   	if (ipaddr == 0) {
	   	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	} else {
		sa.sin_addr.s_addr = ipaddr;
	}

	if (gdisp == -1) { /* If it wasn't specified */
		/* find an acceptable port */
		for (i = XPORT+stdisp; i <= XPORT+endisp; i++) {
			sa.sin_port = htons(i);
			if (bind(s, &sa, sizeof(sa)) == 0) {
				printf("listening on port %d\n", i); break;
			}
			else printf("bind: %d: %s\n", i, sys_errlist[errno]);
		}

	}
	else {
		sa.sin_port = htons(gdisp); 
		if (bind(s, &sa, sizeof(sa)) == -1) perror("bind"), exit(1);
	}

    if (listen(s, 5) == -1) perror("listen"), exit(1);

	return s;
}

int tcpcsocket(host, port)
char *host; int port;
{
	int s, e;
	struct sockaddr_in sa;
	struct hostent *hent;

	if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) exit(1);

	if (! (hent = gethostbyname(host))) {
		e = errno;
		if ((sa.sin_addr.s_addr = inet_addr(host)) == -1)
			fprintf(stderr, "%s: %s\n", host, sys_errlist[e]), exit(1);
	}
   	else bcopy(hent->h_addr, (char *) &sa.sin_addr, hent->h_length);

	sa.sin_family = AF_INET; sa.sin_port = htons(port);

	if (connect(s, &sa, sizeof(sa)) == -1)
		fprintf(stderr, "%s: connect: %s\n", host, sys_errlist[errno]), exit(1);

	return s;
}
